#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>

#include "hashtable.h"

static HashTable nvram;
#define MAX_KEY_SIZE 256
#define MAX_VAL_SIZE 256

int acosNvramConfig_match(char *key, char *value) {
	int res = 0;
	char *key_buf = malloc(MAX_KEY_SIZE);
	memset(key_buf, 0, MAX_KEY_SIZE);
	strcpy(key_buf, key);

	if (ht_contains(&nvram, key_buf)) {
		char *tmp = ht_lookup(&nvram, key_buf);
		res = (0 == strcmp(tmp, value));
	}

	free(key_buf);

	printf("%d = acosNvramConfig_match(%s, %s)\n", res, key, value);
	return res;
}

char* acosNvramConfig_get(char *key) {
	char *res = NULL;
	char *key_buf = malloc(MAX_KEY_SIZE);
	memset(key_buf, 0, MAX_KEY_SIZE);
	strcpy(key_buf, key);

	if (ht_contains(&nvram, key_buf)) {
		res = ht_lookup(&nvram, key_buf);
	}

	free(key_buf);

	printf("%s = acosNvramConfig_get(%s)\n", res, key);
	return res;
}

int send(int sockfd, const void *buf, size_t len, int flags)
{
	return write(sockfd, buf, len);
}

void load_nvram(HashTable* nvram)
{
	char *entry = NULL;
	size_t entry_len = 0;
	char *key_buf = malloc(MAX_KEY_SIZE);
	char *val_buf = malloc(MAX_VAL_SIZE);

	FILE *f = fopen("/etc/nvram", "rb");
	assert(f);
	// seek to first obvious entry
	// this parsing code is probably fragile - good thing we're not production
	fseek(f, 0x20, SEEK_SET);

	while (!feof(f)) {
		// get the nvram entry
		getdelim(&entry, &entry_len, '\0', f);


		// split the entry into kv pairs - must have a key
		char *key = strtok(entry, "=");
		assert(key);

		// val may be null
		char *val = strtok(NULL, "=");

//		printf("key=%s val=%s\n", key, val);

		// fill in the key buffer
		memset(key_buf, 0, MAX_KEY_SIZE);
		strcpy(key_buf, key);

		// fill in the value buffer, if present.
		memset(val_buf, 0, MAX_VAL_SIZE);
		if (val) {
			strcpy(val_buf, val);
		}

		free(entry);

		ht_insert(nvram, key_buf, val_buf);

		entry = NULL;
		entry_len = 0;
	}

	fclose(f);
	free(val_buf);
	free(key_buf);
}

// the real signature is longer but turns out it doesn't matter
void __uClibc_main(void *main, int argc, char** argv)
{
    char req[4096 + 1];
    uint32_t in_addr = 3232235777ul;

    if (argc != 2) {
        printf("Usage: %s <fuzz-file>\n", argv[0]);
        exit(1);
    }

    ht_setup(&nvram, MAX_KEY_SIZE, MAX_VAL_SIZE, 10);
    ht_reserve(&nvram, 100);

    load_nvram(&nvram);

    char *fuzzfile = argv[1];
    int fd = open(fuzzfile, O_RDONLY);

    int n_read = 0; //read(fd, &in_addr, sizeof(in_addr));
    //if (n_read != sizeof(in_addr)) exit(1);

    n_read = read(fd, req, sizeof(req) - 1);
    if (n_read < 0) exit(1);
    req[n_read] = 0;

    fprintf(stderr, "Request: %s\n\n", req);

    // declare a function pointer pointing to the real parse_http_req
    void (*parse_http_req)(char *, void *, uint32_t, int) = (void *)0x409080;

    // Skip a lot of device init and get right to the server setup and http handler
    parse_http_req(req, NULL, in_addr, STDERR_FILENO);

    // need to exit here b/c this function is expected to not return
    exit(0);
}

